from contextvars import ContextVar
import logging
from typing import Any, Callable, List, Literal, Optional, Union

from .middleware.create_bearer_auth import BearerAuthConfig
from .types import AuthInfo, VerifyAccessTokenFunction
from .config import AuthServerConfig, ServerMetadataPaths
from .exceptions import MCPAuthAuthServerException, AuthServerExceptionCode
from .utils import validate_server_config
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import Response, JSONResponse
from starlette.requests import Request
from starlette.routing import Route

_context_var_name = "mcp_auth_context"


class MCPAuth:
    """
    The main class for the mcp-auth library, which provides methods for creating middleware
    functions for handling OAuth 2.0-related tasks and bearer token auth.

    See Also: https://mcp-auth.dev for more information about the library and its usage.
    """

    server: AuthServerConfig
    """
    The configuration for the remote authorization server.
    """

    def __init__(
        self,
        server: AuthServerConfig,
        context_var: ContextVar[Optional[AuthInfo]] = ContextVar(
            _context_var_name, default=None
        ),
    ):
        """
        :param server: Configuration for the remote authorization server.
        :param context_var: Context variable to store the `AuthInfo` object for the current request.
        By default, it will be created with the name "mcp_auth_context".
        """

        result = validate_server_config(server)

        if not result.is_valid:
            logging.error(
                "The authorization server configuration is invalid:\n"
                f"{result.errors}\n"
            )
            raise MCPAuthAuthServerException(
                AuthServerExceptionCode.INVALID_SERVER_CONFIG, cause=result
            )

        if len(result.warnings) > 0:
            logging.warning("The authorization server configuration has warnings:\n")
            for warning in result.warnings:
                logging.warning(f"- {warning}")

        self.server = server
        self._context_var = context_var

    @property
    def auth_info(self) -> Optional[AuthInfo]:
        """
        The current `AuthInfo` object from the context variable.

        This is useful for accessing the authenticated user's information in later middleware or
        route handlers.
        :return: The current `AuthInfo` object, or `None` if not set.
        """

        return self._context_var.get()

    def metadata_endpoint(self) -> Callable[[Request], Any]:
        """
        Returns a Starlette endpoint function that handles the OAuth 2.0 Authorization Metadata
        endpoint (`/.well-known/oauth-authorization-server`) with CORS support.

        Example:
        ```python
        from starlette.applications import Starlette
        from mcpauth import MCPAuth
        from mcpauth.config import ServerMetadataPaths

        mcp_auth = MCPAuth(server=your_server_config)
        app = Starlette(routes=[
            Route(
                ServerMetadataPaths.OAUTH.value,
                mcp_auth.metadata_endpoint(),
                methods=["GET", "OPTIONS"] # Ensure to handle both GET and OPTIONS methods
            )
        ])
        ```
        """

        async def endpoint(request: Request) -> Response:
            if request.method == "OPTIONS":
                response = Response(status_code=204)
            else:
                server_config = self.server
                response = JSONResponse(
                    server_config.metadata.model_dump(exclude_none=True),
                    status_code=200,
                )
            response.headers["Access-Control-Allow-Origin"] = "*"
            response.headers["Access-Control-Allow-Methods"] = "GET, OPTIONS"
            response.headers["Access-Control-Allow-Headers"] = "*"
            return response

        return endpoint

    def metadata_route(self) -> Route:
        """
        Returns a Starlette route that handles the OAuth 2.0 Authorization Metadata endpoint
        (`/.well-known/oauth-authorization-server`) with CORS support.

        Example:
        ```python
        from starlette.applications import Starlette
        from mcpauth import MCPAuth

        mcp_auth = MCPAuth(server=your_server_config)
        app = Starlette(routes=[mcp_auth.metadata_route()])
        ```
        """

        return Route(
            ServerMetadataPaths.OAUTH.value,
            self.metadata_endpoint(),
            methods=["GET", "OPTIONS"],
        )

    def bearer_auth_middleware(
        self,
        mode_or_verify: Union[Literal["jwt"], VerifyAccessTokenFunction],
        audience: Optional[str] = None,
        required_scopes: Optional[List[str]] = None,
        show_error_details: bool = False,
        leeway: float = 60,
    ) -> type[BaseHTTPMiddleware]:
        """
        Creates a middleware that handles bearer token authentication.

        :param mode_or_verify: If "jwt", uses built-in JWT verification; or a custom function that
        takes a string token and returns an `AuthInfo` object.
        :param audience: Optional audience to verify against the token.
        :param required_scopes: Optional list of scopes that the token must contain.
        :param show_error_details: Whether to include detailed error information in the response.
        Defaults to `False`.
        :param leeway: Optional leeway in seconds for JWT verification (`jwt.decode`). Defaults to
        `60`. Not used if a custom function is provided.
        :return: A middleware class that can be used in a Starlette or FastAPI application.
        """

        metadata = self.server.metadata
        if isinstance(mode_or_verify, str) and mode_or_verify == "jwt":
            from .utils import create_verify_jwt

            if not metadata.jwks_uri:
                raise MCPAuthAuthServerException(
                    AuthServerExceptionCode.MISSING_JWKS_URI
                )

            verify = create_verify_jwt(
                metadata.jwks_uri,
                leeway=leeway,
            )
        elif callable(mode_or_verify):
            verify = mode_or_verify
        else:
            raise ValueError(
                "mode_or_verify must be 'jwt' or a callable function that verifies tokens."
            )

        from .middleware.create_bearer_auth import create_bearer_auth

        return create_bearer_auth(
            verify,
            config=BearerAuthConfig(
                issuer=metadata.issuer,
                audience=audience,
                required_scopes=required_scopes,
                show_error_details=show_error_details,
            ),
            context_var=self._context_var,
        )



from enum import Enum
from pydantic import BaseModel
from typing import Any, Optional, List, Dict, Union
from .types import Record

ExceptionCause = Optional[Union[Record, Exception, BaseModel]]


class MCPAuthException(Exception):
    """
    Base class for all mcp-auth exceptions.
    This class provides a standardized way to handle exceptions related to MCP authentication and authorization.
    """

    def __init__(
        self, code: Union[str, Enum], message: str, cause: ExceptionCause = None
    ):
        """
        :param code: A unique exception code in snake_case format that identifies the type of exception.
        :param message: A human-readable description of the exception.
        """

        self.code = code
        self.message = message
        self.cause = cause
        super().__init__(message)

    def to_json(self, show_cause: bool = False) -> Record:
        """
        Converts the exception to a HTTP response friendly JSON format.

        :param show_cause: Whether to include the cause of the exception in the JSON response. Defaults to `false`.
        """
        data: Record = {
            "error": self.code.value if isinstance(self.code, Enum) else self.code,
            "error_description": self.message,
            "cause": (
                (
                    self.cause.model_dump(exclude_none=True)
                    if isinstance(self.cause, BaseModel)
                    else str(self.cause)
                )
                if show_cause and hasattr(self, "cause")
                else None
            ),
        }
        return {k: v for k, v in data.items() if v is not None}


class MCPAuthConfigException(MCPAuthException):
    """
    Exception thrown when there is a configuration issue with mcp-auth.
    """


class AuthServerExceptionCode(str, Enum):
    INVALID_SERVER_METADATA = "invalid_server_metadata"
    INVALID_SERVER_CONFIG = "invalid_server_config"
    MISSING_JWKS_URI = "missing_jwks_uri"


auth_server_exception_description: Dict[AuthServerExceptionCode, str] = {
    AuthServerExceptionCode.INVALID_SERVER_METADATA: "The server metadata is invalid or malformed.",
    AuthServerExceptionCode.INVALID_SERVER_CONFIG: "The server configuration does not match the MCP specification.",
    AuthServerExceptionCode.MISSING_JWKS_URI: "The server metadata does not contain a JWKS URI, which is required for JWT verification.",
}


class MCPAuthAuthServerException(MCPAuthException):
    """
    Exception thrown when there is an issue with the remote authorization server.
    """

    def __init__(self, code: AuthServerExceptionCode, cause: ExceptionCause = None):
        super().__init__(
            code.value,
            auth_server_exception_description.get(
                code, "An exception occurred with the authorization server."
            ),
        )
        self.code = code
        self.cause = cause


class BearerAuthExceptionCode(str, Enum):
    MISSING_AUTH_HEADER = "missing_auth_header"
    INVALID_AUTH_HEADER_FORMAT = "invalid_auth_header_format"
    MISSING_BEARER_TOKEN = "missing_bearer_token"
    INVALID_ISSUER = "invalid_issuer"
    INVALID_AUDIENCE = "invalid_audience"
    MISSING_REQUIRED_SCOPES = "missing_required_scopes"
    INVALID_TOKEN = "invalid_token"


bearer_auth_exception_description: Dict[BearerAuthExceptionCode, str] = {
    BearerAuthExceptionCode.MISSING_AUTH_HEADER: "Missing `Authorization` header. Please provide a valid bearer token.",
    BearerAuthExceptionCode.INVALID_AUTH_HEADER_FORMAT: 'Invalid `Authorization` header format. Expected "Bearer <token>".',
    BearerAuthExceptionCode.MISSING_BEARER_TOKEN: "Missing bearer token in `Authorization` header. Please provide a valid token.",
    BearerAuthExceptionCode.INVALID_ISSUER: "The token issuer does not match the expected issuer.",
    BearerAuthExceptionCode.INVALID_AUDIENCE: "The token audience does not match the expected audience.",
    BearerAuthExceptionCode.MISSING_REQUIRED_SCOPES: "The token does not contain the necessary scopes for this request.",
    BearerAuthExceptionCode.INVALID_TOKEN: "The provided token is not valid or has expired.",
}


class MCPAuthBearerAuthExceptionDetails(BaseModel):
    cause: Any = None
    uri: Optional[str] = None
    missing_scopes: Optional[List[str]] = None
    expected: Any = None
    actual: Any = None


class MCPAuthBearerAuthException(MCPAuthException):
    """
    Exception thrown when there is an issue when authenticating with Bearer tokens.
    """

    def __init__(
        self,
        code: BearerAuthExceptionCode,
        cause: Optional[MCPAuthBearerAuthExceptionDetails] = None,
    ):
        super().__init__(
            code.value,
            bearer_auth_exception_description.get(
                code, "An exception occurred with the Bearer auth."
            ),
        )
        self.code = code
        self.cause = cause

    def to_json(self, show_cause: bool = False) -> Dict[str, Optional[str]]:
        # Matches the OAuth 2.0 exception response format at best effort
        data = super().to_json(show_cause)
        if self.cause:
            data.update(
                {
                    "error_uri": self.cause.uri,
                    "missing_scopes": self.cause.missing_scopes,
                }
            )
        return {k: v for k, v in data.items() if v is not None}


class MCPAuthTokenVerificationExceptionCode(str, Enum):
    INVALID_TOKEN = "invalid_token"
    TOKEN_VERIFICATION_FAILED = "token_verification_failed"


token_verification_exception_description: Dict[
    MCPAuthTokenVerificationExceptionCode, str
] = {
    MCPAuthTokenVerificationExceptionCode.INVALID_TOKEN: "The provided token is invalid or malformed.",
    MCPAuthTokenVerificationExceptionCode.TOKEN_VERIFICATION_FAILED: "The token verification failed due to an error in the verification process.",
}


class MCPAuthTokenVerificationException(MCPAuthException):
    """
    Exception thrown when there is an issue when verifying access tokens.
    """

    def __init__(
        self, code: MCPAuthTokenVerificationExceptionCode, cause: ExceptionCause = None
    ):
        super().__init__(
            code.value,
            token_verification_exception_description.get(
                code, "An exception occurred while verifying the token."
            ),
        )
        self.code = code
        self.cause = cause



from enum import Enum
from typing import List, Optional
from pydantic import BaseModel


class AuthorizationServerMetadata(BaseModel):
    """
    Pydantic model for OAuth 2.0 Authorization Server Metadata as defined in RFC 8414.
    """

    issuer: str
    """
    The authorization server's issuer identifier, which is a URL that uses the `https` scheme and
    has no query or fragment components.
    """

    authorization_endpoint: str
    """
    URL of the authorization server's authorization endpoint [[RFC6749](https://rfc-editor.org/rfc/rfc6749)].
    This is REQUIRED unless no grant types are supported that use the authorization endpoint.

    See: https://rfc-editor.org/rfc/rfc6749#section-3.1
    """

    token_endpoint: str
    """
    URL of the authorization server's token endpoint [[RFC6749](https://rfc-editor.org/rfc/rfc6749)].
    This is REQUIRED unless only the implicit grant type is supported.

    See: https://rfc-editor.org/rfc/rfc6749#section-3.2
    """

    jwks_uri: Optional[str] = None
    """
    URL of the authorization server's JWK Set [[JWK](https://www.rfc-editor.org/rfc/rfc8414.html#ref-JWK)] document.
    The referenced document contains the signing key(s) the client uses to validate signatures 
    from the authorization server. This URL MUST use the `https` scheme.
    """

    registration_endpoint: Optional[str] = None
    """
    URL of the authorization server's OAuth 2.0 Dynamic Client Registration endpoint 
    [[RFC7591](https://www.rfc-editor.org/rfc/rfc7591)].
    """

    scope_supported: Optional[List[str]] = None

    response_types_supported: List[str]
    """
    JSON array containing a list of the OAuth 2.0 `response_type` values that this authorization
    server supports. The array values used are the same as those used with the `response_types`
    parameter defined by "OAuth 2.0 Dynamic Client Registration Protocol" [[RFC7591](https://www.rfc-editor.org/rfc/rfc7591)].
    """

    response_modes_supported: Optional[List[str]] = None
    """
    JSON array containing a list of the OAuth 2.0 `response_mode` values that this
    authorization server supports, as specified in "OAuth 2.0 Multiple Response Type Encoding Practices"
    [[OAuth.Responses](https://datatracker.ietf.org/doc/html/rfc8414#ref-OAuth.Responses)].

    If omitted, the default is ["query", "fragment"]. The response mode value `"form_post"` is
    also defined in "OAuth 2.0 Form Post Response Mode" [[OAuth.FormPost](https://datatracker.ietf.org/doc/html/rfc8414#ref-OAuth.Post)].
    """

    grant_types_supported: Optional[List[str]] = None
    """
    JSON array containing a list of the OAuth 2.0 grant type values that this authorization server supports.
    The array values used are the same as those used with the `grant_types` parameter defined by 
    "OAuth 2.0 Dynamic Client Registration Protocol" [[RFC7591](https://www.rfc-editor.org/rfc/rfc7591)].

    If omitted, the default value is ["authorization_code", "implicit"].
    """

    token_endpoint_auth_methods_supported: Optional[List[str]] = None
    token_endpoint_auth_signing_alg_values_supported: Optional[List[str]] = None
    service_documentation: Optional[str] = None
    ui_locales_supported: Optional[List[str]] = None
    op_policy_uri: Optional[str] = None
    op_tos_uri: Optional[str] = None

    revocation_endpoint: Optional[str] = None
    """
    URL of the authorization server's OAuth 2.0 revocation endpoint [[RFC7009](https://www.rfc-editor.org/rfc/rfc7009)].
    """

    revocation_endpoint_auth_methods_supported: Optional[List[str]] = None
    revocation_endpoint_auth_signing_alg_values_supported: Optional[List[str]] = None

    introspection_endpoint: Optional[str] = None
    """
    URL of the authorization server's OAuth 2.0 introspection endpoint [[RFC7662](https://www.rfc-editor.org/rfc/rfc7662)].
    """

    introspection_endpoint_auth_methods_supported: Optional[List[str]] = None
    introspection_endpoint_auth_signing_alg_values_supported: Optional[List[str]] = None

    code_challenge_methods_supported: Optional[List[str]] = None
    """
    JSON array containing a list of Proof Key for Code Exchange (PKCE) [[RFC7636](https://www.rfc-editor.org/rfc/rfc7636)] 
    code challenge methods supported by this authorization server.
    """

    userinfo_endpoint: Optional[str] = None
    """
    URL of the authorization server's UserInfo endpoint [[OpenID Connect](https://openid.net/specs/openid-connect-core-1_0.html#UserInfo)].
    """


class AuthServerType(str, Enum):
    """
    The type of the authorization server. This information should be provided by the server
    configuration and indicates whether the server is an OAuth 2.0 or OpenID Connect (OIDC)
    authorization server.
    """

    OAUTH = "oauth"
    OIDC = "oidc"


class AuthServerConfig(BaseModel):
    """
    Configuration for the remote authorization server integrated with the MCP server.
    """

    metadata: AuthorizationServerMetadata
    """
    The metadata of the authorization server, which should conform to the MCP specification
    (based on OAuth 2.0 Authorization Server Metadata).

    This metadata is typically fetched from the server's well-known endpoint (OAuth 2.0
    Authorization Server Metadata or OpenID Connect Discovery); it can also be provided
    directly in the configuration if the server does not support such endpoints.

    See:
    - OAuth 2.0 Authorization Server Metadata: https://datatracker.ietf.org/doc/html/rfc8414
    - OpenID Connect Discovery: https://openid.net/specs/openid-connect-discovery-1_0.html
    """

    type: AuthServerType
    """
    The type of the authorization server. See `AuthServerType` for possible values.
    """


class ServerMetadataPaths(str, Enum):
    """
    Enum for server metadata paths.
    This is used to define the standard paths for OAuth and OIDC well-known URLs.
    """

    OAUTH = "/.well-known/oauth-authorization-server"
    OIDC = "/.well-known/openid-configuration"



from typing import Annotated, Dict, List, Optional, Protocol, Union, Any
from pydantic import BaseModel, StringConstraints


Record = Dict[str, Any]


class AuthInfo(BaseModel):
    """
    Authentication information extracted from tokens.

    These fields can be used in the MCP handlers to provide more context about the authenticated
    identity.
    """

    token: str
    """
    The raw access token received in the request. This is typically a JWT or opaque token that is
    used to authenticate the request.
    """

    issuer: str
    """
    The issuer of the access token, which is typically the OAuth / OIDC provider that issued the token.
    This is usually a URL that identifies the authorization server.

    See Also:
    - https://datatracker.ietf.org/doc/html/rfc7519#section-4.1.1
    - https://openid.net/specs/openid-connect-core-1_0.html#IssuerIdentifier
    """

    client_id: Optional[str] = None
    """
    The client ID of the OAuth client that the token was issued to. This is typically the client ID
    registered with the OAuth / OIDC provider.

    Some providers may use 'application ID' or similar terms instead of 'client ID'.
    """

    scopes: List[str] = []
    """
    The scopes (permissions) that the access token has been granted. Scopes define what actions the
    token can perform on behalf of the user or client. Normally, you need to define these scopes in
    the OAuth / OIDC provider and assign them to the `subject` of the token.

    The provider may support different mechanisms for defining and managing scopes, such as
    role-based access control (RBAC) or fine-grained permissions.
    """

    subject: str
    """
    The `sub` (subject) claim of the token, which typically represents the user ID or principal
    that the token is issued for.

    See Also:
    - https://datatracker.ietf.org/doc/html/rfc7519#section-4.1.2
    """

    audience: Optional[Union[str, List[str]]] = None
    """
    The `aud` (audience) claim of the token, which indicates the intended recipient(s) of the token.

    For OAuth / OIDC providers that support Resource Indicators (RFC 8707), this claim can be used
    to specify the intended Resource Server (API) that the token is meant for.

    If the token is intended for multiple audiences, this can be a list of strings.

    See Also:
    - https://datatracker.ietf.org/doc/html/rfc7519#section-4.1.3
    - https://datatracker.ietf.org/doc/html/rfc8707
    """

    claims: Dict[str, Any]
    """
    The raw claims from the token, which can include any additional information provided by the
    token issuer.
    """


class VerifyAccessTokenFunction(Protocol):
    """
    Function type for verifying an access token.

    This function should throw an `MCPAuthTokenVerificationException` if the token is invalid, or return an
    `AuthInfo` instance if the token is valid.

    For example, if you have a JWT verification function, it should at least check the token's
    signature, validate its expiration, and extract the necessary claims to return an `AuthInfo`
    instance.

    Note:
        There's no need to verify the following fields in the token, as they will be checked
        by the MCP handlers:

        - `iss` (issuer)
        - `aud` (audience)
        - `scope` (scopes)
    """

    def __call__(self, token: str) -> AuthInfo:
        """
        :param token: The access token to verify.
        :return: An `AuthInfo` instance containing the extracted authentication information.
        """
        ...


NonEmptyString = Annotated[str, StringConstraints(min_length=1)]


class JwtPayload(BaseModel):
    """
    The base model for JWT (JSON Web Token) payload claims.
    This model defines the common claims that are expected in a JWT used for authentication and
    authorization.
    """

    aud: Optional[Union[NonEmptyString, List[NonEmptyString]]] = None
    """
    The `aud` (audience) claim of the token, which indicates the intended recipient(s) of the token.

    For OAuth / OIDC providers that support Resource Indicators (RFC 8707), this claim can be used
    to specify the intended Resource Server (API) that the token is meant for.

    If the token is intended for multiple audiences, this can be a list of strings.

    See Also:
    - https://datatracker.ietf.org/doc/html/rfc7519#section-4.1.3
    - https://datatracker.ietf.org/doc/html/rfc8707
    """

    iss: NonEmptyString
    """
    The issuer of the access token, which is typically the OAuth / OIDC provider that issued the token.
    This is usually a URL that identifies the authorization server.

    See Also:
    - https://datatracker.ietf.org/doc/html/rfc7519#section-4.1.1
    - https://openid.net/specs/openid-connect-core-1_0.html#IssuerIdentifier
    """

    client_id: NonEmptyString
    """
    The client ID of the OAuth client that the token was issued to. This is typically the client ID
    registered with the OAuth / OIDC provider.

    Some providers may use 'application ID' or similar terms instead of 'client ID'.
    """

    sub: NonEmptyString
    """
    The `sub` (subject) claim of the token, which typically represents the user ID or principal
    that the token is issued for.

    See Also:
    - https://datatracker.ietf.org/doc/html/rfc7519#section-4.1.2
    """

    scope: Optional[Union[str, List[str]]] = None
    """
    The scopes (permissions) that the access token has been granted. Scopes define what actions the
    token can perform on behalf of the user or client. Normally, you need to define these scopes in
    the OAuth / OIDC provider and assign them to the `subject` of the token.

    The provider may support different mechanisms for defining and managing scopes, such as
    role-based access control (RBAC) or fine-grained permissions.
    """

    scopes: Optional[Union[str, List[str]]] = None
    """
    The fallback for the `scope` claim.
    """

    exp: Optional[int] = None
    """
    The expiration time of the access token, represented as a Unix timestamp (seconds since epoch).
    """
